#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "zdbg.h"
#include "smdbgen.h"

const char *usage =
"Usage: smdbgen [OPTIONS] <file>\n\
\n\
Options:\n\
    -h        Print this message and exit.\n\
    -v [0-2]  Specify a level of verbosity. Default: 1\n\
                0 = print number of rooms found, no file output\n\
                1 = output the rom's mdb to \"mdb.txt\"\n\
                2 = output to \"mdb.txt\" and output detailed\n\
                  info on each room header to separate files.\n\
                  All files will be placed in the subdirectory \"mdb\".\n";

int main(int argc, char *argv[])
{
	FILE *rom = NULL;
	mdb_t *mdb = NULL;
	int i = 0;
	int success = 0;
	int verbosity = 1;
	int rom_path = 0;
	int skip = 0;

	if(argc > 1) {
		for(i = 1; i < argc; i++) {
			if(!skip) {
				if(argv[i][0] == '-') {
					switch(argv[i][1]) {
						case 'h':
							printf(usage);
							return 0;

						case 'v':
							if(i == argc - 1) {
								printf("Option requires parameter: -v\n%s", usage);
								return 1;
							}

							if(argv[i + 1][0] < 0x30 || argv[i + 1][0] > 0x39) {
								printf("Invalid parameter for option -v: %s\n%s", argv[i + 1], usage);
								return 1;
							}

							verbosity = atoi(argv[i + 1]);
							if(verbosity < 0 || verbosity > 2) {
								printf("Invalid parameter for option -v: %s\n%s", argv[i + 1], usage);
								return 1;
							}

							skip = 1;
							break;

						default:
							printf("Invalid option: %s\n%s", argv[i], usage);
							return 1;
					}
				} else if(i == argc - 1) {	
					if(access(argv[i], F_OK)) {
						printf("Invalid file path.\n%s", usage);
						return 1;
					} else rom_path = 1;
				} else {
					printf("Invalid argument: %s\n%s", argv[i], usage);
					return 1;
				}
			} else skip = 0;
		}
	} else {
		printf(usage);
		return 0;
	}

	if(!rom_path) {
		printf("Must specify file path.\n%s", usage);
		return 1;
	}

	rom = fopen(argv[argc - 1], "r+");
	check(rom != NULL, "Error opening rom file for editing");

	fseek(rom, 0, SEEK_END);
	if(ftell(rom) % SIZEOF_BANK) {
		printf("Error: the specified file does not appear to be a SNES ROM\n");
		goto error;
	} else if(ftell(rom) < END_8F + SIZEOF_BANK) {
		printf("Error: the specified file does not appear to be a Super Metroid ROM\n");
		goto error;
	}

	mdb = smdbgen(rom, verbosity);
	printf("Number of rooms: %d\n", mdb->roomc);

	fclose(rom);
	mdb_destroy(mdb);
	return 0;

error:

	fclose(rom);
	mdb_destroy(mdb);
	return 1;
}
