#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "zdbg.h"
#include "smdbgen.h"

static const int item_offsets[NUM_CHKS] = {1, 4, 5, 8};
static const long uppers[NUM_CHKS] = {0xA, 0x10, 0x10, 0x10};
static const int list1_sizes[10] = {1, 1, 1, 1, 1, 1, 1, 1, 1, 2};
static const int list2_sizes[14] = {3, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2};
static const char *list1_strings[10] = {"Room index", "Room area", "X on map",
										"Y on map", "W of room", "H of room",
										"Up scroller", "Down scroller", "Special graphics bitflags",
										"Door out pointer"};
static const char *list2_strings[14] = {"Level data pointer for [STANDARD]", "Graphics set", "Music (collection)",
										"Music (play)", "FX1 pointer", "Enemy pop/allowed pointer",
										"Enemy set pointer", "Layer 2 scrolling", "Scroll pointer",
										"RoomVar (unused)", "FX2 pointer", "PLM pointer",
										"BG_Data pointer", "Setup code pointer"};
static const int states[8] = {0xE5E6, 0xE5FF, 0xE612, 0xE629, 0xE640, 0xE652, 0xE669, 0xE678};
static const char *state_strings[8] =  {"[STANDARD]", "[TOURIAN BOSS 1]", "[EVENTS]",
										"[BOSSES]", "[MORPH]", "[MORPH & MISSILES]",
										"[POWER BOMBS]", "[SPEED BOOSTER]"};
const unsigned char whitespace = 0x20;


unsigned int wtoi(unsigned char *word)
{
	unsigned int word_val = (0x100 * word[1]) + word[0];
	return word_val;
}

unsigned int lwtoi(unsigned char *longword)
{
	unsigned int longword_val = (0x10000 * longword[2]) + (0x100 * longword[1]) + longword[0];
	return longword_val;
}

unsigned char *itow(unsigned int val)
{
	unsigned char *word = malloc(2);
	word[0] = (unsigned char)(val & 0x00FF);
	word[1] = (unsigned char)((val & 0xFF00) / 0x100);

	return word;
}

unsigned char *itolw(unsigned int val)
{
	unsigned char *longword = malloc(3);
	longword[0] = (unsigned char)(val & 0x0000FF);
	longword[1] = (unsigned char)((val & 0x00FF00) / 0x100);
	longword[2] = (unsigned char)((val & 0xFF0000) / 0x10000);

	return longword;
}

unsigned int snes2pc(unsigned int addr)
{
	if(addr < 0x808000 || addr > 0xFFFFFF)
		return -1;

	unsigned int result = ((((addr - (addr & 0x00FFFF)) / 0x10000) - 0x81) * SIZEOF_BANK + (addr & 0x00FFFF)) & 0x00FFFFFF;
	return result;
}

unsigned int pc2snes(unsigned int addr)
{
	if(addr > 0x3FFFFF)
		return -1;

	unsigned int result = 0x808000 + ((addr - (addr % 0x8000)) * 2) + (addr % 0x8000);
	return result;
}

int scan(FILE *rom, int offset)
{
	unsigned char *header = malloc(STATIC_HEADER_SIZE);
	fseek(rom, offset, SEEK_SET);
	fread(header, 1, STATIC_HEADER_SIZE, rom);

	int i = 0;
	int success = 1;
	int success2 = 0;
    for(i = 0; i < NUM_CHKS; i++) {
        if(header[item_offsets[i]] > uppers[i])
            success = 0;
    }

    if(header[STATIC_HEADER_SIZE - 1] == 0xE6 || header[STATIC_HEADER_SIZE - 1] == 0xE5)
    	success2 = 1;

	if(success && success2) {
		free(header);
		return 1;
	} else {
		free(header);
		return 0;
	}
}

mdb_t *mdb_create(int roomc)
{
	if(roomc < 1)
		roomc = 1;
	int i = 0;
	mdb_t *t = malloc(sizeof(mdb_t));
	t->roomc = roomc;
	t->roomv = malloc(roomc * sizeof(unsigned int));
	t->hsizes = malloc(roomc * sizeof(unsigned int));
	for(i = 0; i < roomc; i++) {
		t->roomv[i] = 0;
		t->hsizes[i] = 0;
	}

	return t;
}

void mdb_destroy(mdb_t *mdb)
{
	if(mdb) {
		free(mdb->hsizes);
		free(mdb->roomv);
		free(mdb);
	}
}

int follow(FILE *rom, int offset, int bank)
{
	unsigned char word[2] = {0, 0};
	fseek(rom, offset, SEEK_SET);
	if(!isptr(rom))
		return 1;
	fseek(rom, -2, SEEK_CUR);
	fread(word, 1, 2, rom);
	fseek(rom, bank + wtoi(word) - BEGIN_BANK, SEEK_SET);
	return 0;
}

int isptr(FILE *rom)
{
	unsigned char word[2] = {0, 0};
	fread(word, 1, 2, rom);
	if(wtoi(word) < BEGIN_BANK)
		return 0;
	else
		return 1;
}

int isdoor(FILE *rom)
{
	unsigned char word[2] = {0, 0};
	follow(rom, ftell(rom), BEGIN_83);
	if(follow(rom, ftell(rom), BEGIN_8F)) {
		fseek(rom, -2, SEEK_CUR);
		fread(word, 1, 2, rom);
		if(wtoi(word) == 0)
			return -1;
		else
			return 0;
	}

	int offset = ftell(rom);
	if(!scan(rom, offset))
		return 0;

	return 1;
}

int calc_hsize(FILE *rom, int offset)
{
	int success = 0;
	unsigned char word[2] = {0, 0};
	fseek(rom, offset + 11, SEEK_SET);
	while(!success) {
		fread(word, 1, 2, rom);
		switch(wtoi(word)) {
			case 0xE5E6:
				success = 1;
				break;

			case 0xE5FF:
				break;

			case 0xE612: // events (takes extra byte)
				fseek(rom, 1, SEEK_CUR);
				break;

			case 0xE629: // bosses (takes extra byte)
				fseek(rom, 1, SEEK_CUR);
				break;

			case 0xE640:
				break;

			case 0xE652:
				break;

			case 0xE669:
				break;

			case 0xE678:
				break;

			default:
				fseek(rom, offset, SEEK_SET);
				return -1;
		}

		if(!success)
			fseek(rom, 2, SEEK_CUR);
	}

	return ftell(rom) - offset + 26;
}

void detail_header(FILE *rom, FILE *output, mdb_t *mdb, int index)
{
	int i = 0;
	int j = 0;
	int k = 0;
	int m = 0;
	int n = 0;
	int state_num = 0;
	unsigned int word_val = 0;
	int success = 0;
	int size = mdb->hsizes[index];
	if(size < 1)
		return;
	int rows = (size - size % 16) / 16 + 1;
	unsigned int addr = pc2snes(mdb->roomv[index]);

	unsigned char *header = malloc(size);
	unsigned char word[2] = {0, 0};
	unsigned char *longword = NULL;
	char *strbuf = malloc(50);
	for(i = 0; i < 50; i++) {
		strbuf[i] = 0;
	}

	fseek(rom, mdb->roomv[index], SEEK_SET);
	fread(header, 1, size, rom);
	longword = itolw(addr);
	sprintf(strbuf, "$%02X:%02X%02X [%X room header - 0x%02X/%02d bytes]",
			longword[2], longword[1],
			longword[0], mdb->roomv[index],
			mdb->hsizes[index], mdb->hsizes[index]);

	for(i = 0; i < 50; i++) {
		if(strbuf[i] != 0)
			fputc(strbuf[i], output);
	}

	ENDL(output);
	for(i = 0; i < 50; i++) {
		strbuf[i] = '=';
	}

	fwrite(strbuf, 1, 50, output);
	ENDL(output);
	for(i = 0; i < rows; i++) {
		for(j = 0; j < 16; j++) {
			if(k < size) {
				sprintf(strbuf, "%02X", header[k]);
				fwrite(strbuf, 1, 2, output);
				if(!((j + 1) % 4) && j != 15 && k != size - 1) {
					fputc(whitespace, output);
					fputc(whitespace, output);
				} else if(j != 15 && k != size - 1)
					fputc(whitespace, output);
			}

			k++;
		}

		ENDL(output);
	}

	for(i = 0; i < 50; i++) {
        strbuf[i] = '=';
    }

    fwrite(strbuf, 1, 50, output);
    ENDL(output);

	for(i = 0; i < 10; i++) {
		for(k = 0; k < 3 - list1_sizes[i]; k++) {
			for(m = 0; m < 3; m++) {
				fputc(whitespace, output);
			}
		}

		for(j = 0; j < list1_sizes[i]; j++) {
			sprintf(strbuf, "%02X ", header[n + j]);
			fwrite(strbuf, 1, 3, output);
		}

		n += list1_sizes[i];
		fputc('.', output);
		fputc(' ', output);
		fputs(list1_strings[i], output);
		ENDL(output);
	}

	while(!success) {
		word[0] = header[n];
		word[1] = header[n + 1];
		word_val = wtoi(word);
		for(i = 0; i < 8; i++) {
			if(word_val == states[i]) {
				if(!(word_val == 0xE612 || word_val == 0xE629)) {
					for(j = 0; j < 3; j++) {
						fputc(whitespace, output);
					}
					sprintf(strbuf, "%02X %02X ", header[n], header[n + 1]);
					fwrite(strbuf, 1, 6, output);
					fputc('.', output);
					fputc(' ', output);
					fputs("Room state ", output);
					fputs(state_strings[i], output);
					ENDL(output);
					n += 2;
				} else {
					sprintf(strbuf, "%02X %02X %02X ", header[n], header[n + 1], header[n + 2]);
					fwrite(strbuf, 1, 9, output);
					fputc('.', output);
					fputc(' ', output);
					fputs("Room state ", output);
					fputs(state_strings[i], output);
					ENDL(output);
					n += 3;
				}

				if(states[i] == 0xE5E6)
					success = 1;
				state_num = i;
			}
		}

		if(!success) {
			for(j = 0; j < 3; j++) {
				fputc(whitespace, output);
			}
			sprintf(strbuf, "%02X %02X ", header[n], header[n + 1]);
			fwrite(strbuf, 1, 6, output);
			fputc('.', output);
			fputc(' ', output);
			fputs("Pointer to level data pointer for ", output);
			fputs(state_strings[state_num], output);
			ENDL(output);
			n += 2;
		}
	}

	for(i = 0; i < 14; i++) {
		for(k = 0; k < 3 - list2_sizes[i]; k++) {
			for(m = 0; m < 3; m++) {
				fputc(whitespace, output);
			}
		}

		for(j = 0; j < list2_sizes[i]; j++) {
			sprintf(strbuf, "%02X ", header[n + j]);
			fwrite(strbuf, 1, 3, output);
		}

		n += list2_sizes[i];
		fputc('.', output);
		fputc(' ', output);
		fputs(list2_strings[i], output);
		ENDL(output);
	}

	free(header);
	free(longword);
	free(strbuf);
}

mdb_t *smdbgen(FILE *rom, int verbosity)
{
	mdb_t *mdb = mdb_create(500);
	mdb_t *mdb_ = mdb_create(500);

	check(rom != NULL, "smdbgen() passed NULL rom file stream");

	int i = 0;
	int j = 0;
	int n = 0;
	int rc = 0;
	int doors = 0;
	int success = 0;
	int counter = 0;
	int offset = 0;
	int door_offset;

	for(offset = BEGIN_8F; offset < END_8F; offset++) {
		if(scan(rom, offset)) {
			if(counter < mdb->roomc)
				mdb->roomv[counter] = offset;
			else
				printf("Warning: reached internal room limit (%d)\n", mdb->roomc);
			counter++;
		}
	}

	for(i = 0; i < counter - 1 && i < mdb->roomc - 1; i++) {
		if(mdb->roomv[i + 1] - mdb->roomv[i] < 30) {
			printf("Found near-match entry: %X\n", mdb->roomv[i]);
			mdb->roomv[i] = 0;
		}
	}

	for(i = 0; i < counter; i++) {
		if(mdb->roomv[i]) {
			mdb_->roomv[j] = mdb->roomv[i];
			j++;
		}
	}

	counter = j;
	mdb_destroy(mdb);
	mdb = mdb_create(500);
	for(i = 0; i < counter; i++) {
		mdb->roomv[i] = mdb_->roomv[i];
	}

	mdb_destroy(mdb_);
	mdb->roomc = counter;
	printf("Found %d room header matches\n", counter);
	for(i = 0; i < counter; i++) {
		if(!follow(rom, mdb->roomv[i] + DOOR_OUT, BEGIN_8F)) {
			door_offset = ftell(rom); // offset at which door 00 pointer exists
			n = 0;
			doors = 0;
			success = 0;
			while(!success) {
				if(!isptr(rom))
					success = 1;
				else
					n++;
			}

			if(n) {
				for(j = 0; j < n; j++) {
					fseek(rom, door_offset + (2 * j), SEEK_SET);
					rc = isdoor(rom);
					if(rc == 1)
						doors++;
					else if(rc == -1) {
						if(n > 1)
							doors++;
					}
				}

				if(!doors) {
					printf("[INFO] Room %X is being excluded due to lack of working doors.\n", mdb->roomv[i]);
					mdb->roomv[i] = 0;
					counter--;
				} else if(n - doors > 1)
					printf("[INFO] Room %X has %d/%d malfunctioning doors, including anyway.\n", mdb->roomv[i], n - doors, n);
			} else {
				printf("[INFO] Room %X is being excluded due to invalid door pointers.\n", mdb->roomv[i]);
				mdb->roomv[i] = 0;
				counter--;
			}
		} else {
			printf("[INFO] Room %X is being excluded due to invalid door_out pointer.\n", mdb->roomv[i]);
			mdb->roomv[i] = 0;
			counter--;
		}
	}

	mdb->roomc = counter;
	mdb_ = mdb_create(500);
	mdb_->roomc = mdb->roomc;
	j = 0;
	for(i = 0; i < 500; i++) {
		if(mdb->roomv[i]) {
			mdb_->roomv[j] = mdb->roomv[i];
			j++;
		}
	}

	mdb_destroy(mdb);

	if(verbosity) {
		FILE *output = NULL;
		char *filename = NULL;
		char *fnamebuf = NULL;
		char *strbuf = malloc(5);
		struct stat st = {0};

		if(verbosity == 1) {
			filename = malloc(3 + 4 + 1);
			strcpy(filename, "mdb.txt");
		} else {
			filename = malloc(4 + 3 + 4 + 1);
			strcpy(filename, "mdb/mdb.txt");
			struct stat st = {0};
			if(stat("mdb", &st) == -1)
				mkdir("mdb", 0777);
			fnamebuf = malloc(strlen("mdb/00000.txt") + 1);
		}

		if(!access(filename, F_OK))
			remove(filename);
		output = fopen(filename, "a");
		if(output == NULL) {
			printf("Error opening output file for editing\n");
			free(filename);
			free(fnamebuf);
			free(strbuf);
			goto error;
		}

		for(i = 0; i < counter; i++) {
			if(mdb_->roomv[i]) {
				sprintf(strbuf, "%X", mdb_->roomv[i]);
            	fwrite(strbuf, 1, 5, output);
				ENDL(output);
			}
		}

		printf("Successfully output to file: %s\n", filename);
		fclose(output);
		if(verbosity == 2) {
			for(i = 0; i < counter; i++) {
				sprintf(fnamebuf, "mdb/%X.txt", mdb_->roomv[i]);
				output = fopen(fnamebuf, "w");
				mdb_->hsizes[i] = calc_hsize(rom, mdb_->roomv[i]);
				if(mdb_->hsizes[i] != -1)
					detail_header(rom, output, mdb_, i);
				else
					printf("Unable to detail room %X: detected unknown room state header\n", mdb_->roomv[i]);
				fclose(output);
			}
		}

		free(filename);
		free(fnamebuf);
		free(strbuf);
	}

	return mdb_;

error:

	mdb_destroy(mdb);
	return NULL;
}
